# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.numeric_types as nt
import pyomo.common.unittest as unittest

from pyomo.common.dependencies import numpy, numpy_available
from pyomo.core.expr import LinearExpression
from pyomo.environ import Var

_type_sets = (
    'native_types',
    'native_numeric_types',
    'native_logical_types',
    'native_integer_types',
    'native_complex_types',
)


class TestNativeTypes(unittest.TestCase):
    def setUp(self):
        bool(numpy_available)
        for s in _type_sets:
            setattr(self, s, set(getattr(nt, s)))
            getattr(nt, s).clear()

    def tearDown(self):
        for s in _type_sets:
            getattr(nt, s).clear()
            getattr(nt, s).update(getattr(self, s))

    def test_check_if_native_type(self):
        self.assertEqual(nt.native_types, set())
        self.assertEqual(nt.native_logical_types, set())
        self.assertEqual(nt.native_numeric_types, set())
        self.assertEqual(nt.native_integer_types, set())
        self.assertEqual(nt.native_complex_types, set())

        self.assertTrue(nt.check_if_native_type("a"))
        self.assertIn(str, nt.native_types)
        self.assertNotIn(str, nt.native_logical_types)
        self.assertNotIn(str, nt.native_numeric_types)
        self.assertNotIn(str, nt.native_integer_types)
        self.assertNotIn(str, nt.native_complex_types)

        self.assertTrue(nt.check_if_native_type(1))
        self.assertIn(int, nt.native_types)
        self.assertNotIn(int, nt.native_logical_types)
        self.assertIn(int, nt.native_numeric_types)
        self.assertIn(int, nt.native_integer_types)
        self.assertNotIn(int, nt.native_complex_types)

        self.assertTrue(nt.check_if_native_type(1.5))
        self.assertIn(float, nt.native_types)
        self.assertNotIn(float, nt.native_logical_types)
        self.assertIn(float, nt.native_numeric_types)
        self.assertNotIn(float, nt.native_integer_types)
        self.assertNotIn(float, nt.native_complex_types)

        self.assertTrue(nt.check_if_native_type(True))
        self.assertIn(bool, nt.native_types)
        self.assertIn(bool, nt.native_logical_types)
        self.assertNotIn(bool, nt.native_numeric_types)
        self.assertNotIn(bool, nt.native_integer_types)
        self.assertNotIn(bool, nt.native_complex_types)

        self.assertFalse(nt.check_if_native_type(slice(None, None, None)))
        self.assertNotIn(slice, nt.native_types)
        self.assertNotIn(slice, nt.native_logical_types)
        self.assertNotIn(slice, nt.native_numeric_types)
        self.assertNotIn(slice, nt.native_integer_types)
        self.assertNotIn(slice, nt.native_complex_types)

    def test_check_if_logical_type(self):
        self.assertEqual(nt.native_types, set())
        self.assertEqual(nt.native_logical_types, set())
        self.assertEqual(nt.native_numeric_types, set())
        self.assertEqual(nt.native_integer_types, set())
        self.assertEqual(nt.native_complex_types, set())

        self.assertFalse(nt.check_if_logical_type("a"))
        self.assertNotIn(str, nt.native_types)
        self.assertNotIn(str, nt.native_logical_types)
        self.assertNotIn(str, nt.native_numeric_types)
        self.assertNotIn(str, nt.native_integer_types)
        self.assertNotIn(str, nt.native_complex_types)

        self.assertFalse(nt.check_if_logical_type("a"))

        self.assertTrue(nt.check_if_logical_type(True))
        self.assertIn(bool, nt.native_types)
        self.assertIn(bool, nt.native_logical_types)
        self.assertNotIn(bool, nt.native_numeric_types)
        self.assertNotIn(bool, nt.native_integer_types)
        self.assertNotIn(bool, nt.native_complex_types)

        self.assertTrue(nt.check_if_logical_type(True))

        self.assertFalse(nt.check_if_logical_type(1))
        self.assertNotIn(int, nt.native_types)
        self.assertNotIn(int, nt.native_logical_types)
        self.assertNotIn(int, nt.native_numeric_types)
        self.assertNotIn(int, nt.native_integer_types)
        self.assertNotIn(int, nt.native_complex_types)

        if numpy_available:
            self.assertTrue(nt.check_if_logical_type(numpy.bool_(1)))
            self.assertIn(numpy.bool_, nt.native_types)
            self.assertIn(numpy.bool_, nt.native_logical_types)
            self.assertNotIn(numpy.bool_, nt.native_numeric_types)
            self.assertNotIn(numpy.bool_, nt.native_integer_types)
            self.assertNotIn(numpy.bool_, nt.native_complex_types)

    def test_check_if_numeric_type(self):
        self.assertEqual(nt.native_types, set())
        self.assertEqual(nt.native_logical_types, set())
        self.assertEqual(nt.native_numeric_types, set())
        self.assertEqual(nt.native_integer_types, set())
        self.assertEqual(nt.native_complex_types, set())

        self.assertFalse(nt.check_if_numeric_type("a"))
        self.assertFalse(nt.check_if_numeric_type("a"))
        self.assertNotIn(str, nt.native_types)
        self.assertNotIn(str, nt.native_logical_types)
        self.assertNotIn(str, nt.native_numeric_types)
        self.assertNotIn(str, nt.native_integer_types)
        self.assertNotIn(str, nt.native_complex_types)

        self.assertFalse(nt.check_if_numeric_type(True))
        self.assertFalse(nt.check_if_numeric_type(True))
        self.assertNotIn(bool, nt.native_types)
        self.assertNotIn(bool, nt.native_logical_types)
        self.assertNotIn(bool, nt.native_numeric_types)
        self.assertNotIn(bool, nt.native_integer_types)
        self.assertNotIn(bool, nt.native_complex_types)

        self.assertTrue(nt.check_if_numeric_type(1))
        self.assertTrue(nt.check_if_numeric_type(1))
        self.assertIn(int, nt.native_types)
        self.assertNotIn(int, nt.native_logical_types)
        self.assertIn(int, nt.native_numeric_types)
        self.assertIn(int, nt.native_integer_types)
        self.assertNotIn(int, nt.native_complex_types)

        self.assertTrue(nt.check_if_numeric_type(1.5))
        self.assertTrue(nt.check_if_numeric_type(1.5))
        self.assertIn(float, nt.native_types)
        self.assertNotIn(float, nt.native_logical_types)
        self.assertIn(float, nt.native_numeric_types)
        self.assertNotIn(float, nt.native_integer_types)
        self.assertNotIn(float, nt.native_complex_types)

        self.assertFalse(nt.check_if_numeric_type(1j))
        self.assertIn(complex, nt.native_types)
        self.assertNotIn(complex, nt.native_logical_types)
        self.assertNotIn(complex, nt.native_numeric_types)
        self.assertNotIn(complex, nt.native_integer_types)
        self.assertIn(complex, nt.native_complex_types)

        v = Var()
        v.construct()
        self.assertFalse(nt.check_if_numeric_type(v))
        self.assertNotIn(type(v), nt.native_types)
        self.assertNotIn(type(v), nt.native_logical_types)
        self.assertNotIn(type(v), nt.native_numeric_types)
        self.assertNotIn(type(v), nt.native_integer_types)
        self.assertNotIn(type(v), nt.native_complex_types)

        e = LinearExpression([1])
        self.assertFalse(nt.check_if_numeric_type(e))
        self.assertNotIn(type(e), nt.native_types)
        self.assertNotIn(type(e), nt.native_logical_types)
        self.assertNotIn(type(e), nt.native_numeric_types)
        self.assertNotIn(type(e), nt.native_integer_types)
        self.assertNotIn(type(e), nt.native_complex_types)

        if numpy_available:
            self.assertFalse(nt.check_if_numeric_type(numpy.bool_(1)))
            self.assertNotIn(numpy.bool_, nt.native_types)
            self.assertNotIn(numpy.bool_, nt.native_logical_types)
            self.assertNotIn(numpy.bool_, nt.native_numeric_types)
            self.assertNotIn(numpy.bool_, nt.native_integer_types)
            self.assertNotIn(numpy.bool_, nt.native_complex_types)

            self.assertFalse(nt.check_if_numeric_type(numpy.array([1])))
            self.assertNotIn(numpy.ndarray, nt.native_types)
            self.assertNotIn(numpy.ndarray, nt.native_logical_types)
            self.assertNotIn(numpy.ndarray, nt.native_numeric_types)
            self.assertNotIn(numpy.ndarray, nt.native_integer_types)
            self.assertNotIn(numpy.ndarray, nt.native_complex_types)

            self.assertTrue(nt.check_if_numeric_type(numpy.float64(1)))
            self.assertIn(numpy.float64, nt.native_types)
            self.assertNotIn(numpy.float64, nt.native_logical_types)
            self.assertIn(numpy.float64, nt.native_numeric_types)
            self.assertNotIn(numpy.float64, nt.native_integer_types)
            self.assertNotIn(numpy.float64, nt.native_complex_types)

            self.assertTrue(nt.check_if_numeric_type(numpy.int64(1)))
            self.assertIn(numpy.int64, nt.native_types)
            self.assertNotIn(numpy.int64, nt.native_logical_types)
            self.assertIn(numpy.int64, nt.native_numeric_types)
            self.assertIn(numpy.int64, nt.native_integer_types)
            self.assertNotIn(numpy.int64, nt.native_complex_types)

            self.assertFalse(nt.check_if_numeric_type(numpy.complex128(1)))
            self.assertIn(numpy.complex128, nt.native_types)
            self.assertNotIn(numpy.complex128, nt.native_logical_types)
            self.assertNotIn(numpy.complex128, nt.native_numeric_types)
            self.assertNotIn(numpy.complex128, nt.native_integer_types)
            self.assertIn(numpy.complex128, nt.native_complex_types)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Test data model tree for QTreeView.  These tests need PyQt.
"""

import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Var,
    BooleanVar,
    Constraint,
    Objective,
    Reals,
    Block,
    Expression,
    ExternalFunction,
    sin,
    sqrt,
)
import pyomo.environ as pyo
from pyomo.contrib.viewer.model_browser import ComponentDataModel
import pyomo.contrib.viewer.qt as myqt
from pyomo.common.dependencies import DeferredImportError
from pyomo.core.base.units_container import pint_available

available = myqt.available

if available:
    from pyomo.contrib.viewer.ui_data import UIData
    import pyomo.contrib.viewer.ui as ui
else:

    class UIData:
        model = None

        def __init__(*args, **kwargs):
            pass


@unittest.skipIf(not available or not pint_available, "PyQt or units not available")
class TestDataModel(unittest.TestCase):
    def setUp(self):
        # Borrowed this test model from the trust region tests
        m = ConcreteModel(name="tm")
        m.y = BooleanVar(range(3), initialize=False)
        m.z = Var(
            range(3), domain=Reals, initialize=2.0, units=pyo.units.m, doc="test doc"
        )
        m.x = Var(range(2), initialize=2.0, units=pyo.units.m)
        m.x[1] = 1.0

        m.b1 = Block()
        m.b1.e1 = Expression(expr=m.x[0] + m.x[1])

        def blackbox(a, b):
            return sin(a - b)

        self.bb = ExternalFunction(blackbox)

        m.obj = Objective(
            expr=(m.z[0] - 1.0) ** 2
            + (m.z[0] - m.z[1]) ** 2
            + (m.z[2] - 1.0) ** 2
            + (m.x[0] - 1.0) ** 4
            + (m.x[1] - 1.0) ** 6  # + m.bb(m.x[0],m.x[1])
        )
        m.c1 = Constraint(
            expr=m.x[0] * m.z[0] ** 2 + self.bb(m.x[0], m.x[1]) == 2 * sqrt(2.0)
        )
        m.c2 = Constraint(expr=m.z[2] ** 4 * m.z[1] ** 2 + m.z[1] == 8 + sqrt(2.0))

        # Add a few items to test more
        m.b2 = Block([1, 2])
        m.b2[1].w = Var(initialize=4)
        m.b2[2].w = Var(initialize=4)
        self.m = m.clone()

    def test_create_tree_var(self):
        ui_data = UIData(model=self.m)
        # Defaults to variables and two columns name and value
        data_model = ComponentDataModel(parent=None, ui_data=ui_data)
        # There should be one root item
        assert len(data_model.rootItems) == 1
        assert data_model.rootItems[0].data == self.m
        # The children should be in the model construction order,
        # and the indexes are sorted
        children = data_model.rootItems[0].children
        assert children[0].data == self.m.y
        assert children[1].data == self.m.z
        assert children[2].data == self.m.x
        assert children[3].data == self.m.b1
        # Check the data display role The rows in the tree should be:
        #   0. Model
        #     0. y
        #       0. y[0], False
        #       1. y[1], False
        #       2. y[2], False
        #     1. z
        #       0. z[0], 2
        #       1. z[1], 2
        #       2. z[2], 2
        #     2. x
        #       0. x[0], 2
        #       1. x[1], 1
        #     3. b1
        root_index = data_model.index(0, 0)
        assert data_model.data(root_index) == "tm"
        zidx = data_model.index(0, 0, parent=root_index)
        assert data_model.data(zidx) == "y"
        zidx = data_model.index(1, 0, parent=root_index)
        assert data_model.data(zidx) == "z"
        xidx = data_model.index(2, 0, parent=root_index)
        assert data_model.data(xidx) == "x"
        b1idx = data_model.index(3, 0, parent=root_index)
        assert data_model.data(b1idx) == "b1"
        idx = data_model.index(0, 0, parent=zidx)
        assert data_model.data(idx) == "z[0]"
        idx = data_model.index(0, 1, parent=zidx)
        assert abs(data_model.data(idx) - 2.0) < 0.0001
        idx = data_model.index(1, 0, parent=zidx)
        assert data_model.data(idx) == "z[1]"
        idx = data_model.index(1, 1, parent=zidx)
        assert abs(data_model.data(idx) - 2.0) < 0.0001
        # tooltip gives doc
        assert "test doc" == data_model.data(
            zidx, role=myqt.Qt.ItemDataRole.ToolTipRole
        )
        assert myqt.QtCore.Qt.blue == data_model.data(
            zidx, role=myqt.Qt.ItemDataRole.ForegroundRole
        )
        assert myqt.QtCore.Qt.black == data_model.data(
            root_index, role=myqt.Qt.ItemDataRole.ForegroundRole
        )

    def test_create_tree_con(self):
        ui_data = UIData(model=self.m)
        # Make a tree with constraints
        data_model = ComponentDataModel(
            parent=None,
            ui_data=ui_data,
            components=(Constraint,),
            columns=["name", "active"],
        )
        # There should be one root item
        assert len(data_model.rootItems) == 1
        assert data_model.rootItems[0].data == self.m
        # The children should be in the model construction order,
        # and the indexes are sorted
        children = data_model.rootItems[0].children
        assert children[0].data == self.m.b1
        assert children[1].data == self.m.c1
        assert children[2].data == self.m.c2
        # Check the data display role The rows in the tree should be:
        #   0. Model
        #     0. c1, True
        #     1. c2, True
        #     2. b1, True
        root_index = data_model.index(0, 0)
        assert data_model.data(root_index) == "tm"
        idx = data_model.index(0, 0, parent=root_index)
        assert data_model.data(idx) == "b1"
        idx = data_model.index(0, 1, parent=root_index)
        assert data_model.data(idx) == True
        idx = data_model.index(1, 0, parent=root_index)
        assert data_model.data(idx) == "c1"
        idx = data_model.index(2, 0, parent=root_index)
        assert data_model.data(idx) == "c2"
        c2 = idx.internalPointer()
        c2.set("active", False)
        assert not self.m.c2.active
        c2.set("active", True)
        assert self.m.c2.active
        assert "z[2]" in data_model.data(idx, role=myqt.Qt.ItemDataRole.ToolTipRole)

    def test_create_tree_expr(self):
        ui_data = UIData(model=self.m)
        # Make a tree with constraints
        data_model = ComponentDataModel(
            parent=None,
            ui_data=ui_data,
            components=(Expression,),
            columns=["name", "value"],
        )
        # There should be one root item
        assert len(data_model.rootItems) == 1
        assert data_model.rootItems[0].data == self.m
        # The children should be in the model construction order,
        # and the indexes are sorted
        children = data_model.rootItems[0].children
        assert children[0].data == self.m.b1
        assert children[0].children[0].data == self.m.b1.e1
        ui_data.calculate_expressions()
        # Check the data display role The rows in the tree should be:
        #   0. Model
        #     0. b1,
        #       0. e1, value
        root_index = data_model.index(0, 0)
        b1_index = data_model.index(0, 0, parent=root_index)
        e1_index0 = data_model.index(0, 0, parent=b1_index)
        e1_index1 = data_model.index(0, 1, parent=b1_index)
        assert data_model.data(e1_index0) == "b1.e1"
        assert abs(data_model.data(e1_index1) - 3.0) < 0.0001

    def test_update_tree_expr(self):
        ui_data = UIData(model=self.m)
        # Make a tree with constraints
        data_model = ComponentDataModel(
            parent=None,
            ui_data=ui_data,
            components=(Expression,),
            columns=["name", "value"],
        )

        self.m.newe = Expression(expr=self.m.x[0] + self.m.x[1])

        data_model._update_tree()

        # There should be one root item
        assert len(data_model.rootItems) == 1
        assert data_model.rootItems[0].data == self.m
        # The children should be in the model construction order,
        # and the indexes are sorted
        children = data_model.rootItems[0].children
        assert children[0].data == self.m.b1
        assert children[0].children[0].data == self.m.b1.e1
        ui_data.calculate_expressions()
        # Check the data display role The rows in the tree should be:
        #   0. Model
        #     0. b1,
        #       0. e1, value
        root_index = data_model.index(0, 0)
        b1_index = data_model.index(0, 0, parent=root_index)
        e1_index0 = data_model.index(0, 0, parent=b1_index)
        e1_index1 = data_model.index(0, 1, parent=b1_index)
        assert data_model.data(e1_index0) == "b1.e1"
        assert abs(data_model.data(e1_index1) - 3.0) < 0.0001
        # Check that in the update the new expression was added
        found = False
        for i in children:
            if id(i.data) == id(self.m.newe):
                found = True
                break
        assert found



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.contrib.cp import IntervalVar
from pyomo.core import ModelComponentFactory
from pyomo.core.base.component import ActiveComponentData
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import ActiveIndexedComponent
from pyomo.core.base.initializer import Initializer

import sys
from weakref import ref as weakref_ref

logger = logging.getLogger(__name__)


class SequenceVarData(ActiveComponentData):
    """This class defines the abstract interface for a single sequence variable."""

    __slots__ = ('interval_vars',)

    def __init__(self, component=None):
        # in-lining ActiveComponentData and ComponentData constructors, as is
        # traditional:
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET
        self._active = True

        # This thing is really just an ordered set of interval vars that we can
        # write constraints over.
        self.interval_vars = []

    def set_value(self, expr):
        # We'll demand expr be a list for now--it needs to be ordered so this
        # doesn't seem like too much to ask
        if not hasattr(expr, '__iter__'):
            raise ValueError(
                "'expr' for SequenceVar must be a list of IntervalVars. "
                "Encountered type '%s' constructing '%s'" % (type(expr), self.name)
            )
        for v in expr:
            if not hasattr(v, 'ctype') or v.ctype is not IntervalVar:
                raise ValueError(
                    "The SequenceVar 'expr' argument must be a list of "
                    "IntervalVars. The 'expr' for SequenceVar '%s' included "
                    "an object of type '%s'" % (self.name, type(v))
                )
            self.interval_vars.append(v)


@ModelComponentFactory.register("Sequences of IntervalVars")
class SequenceVar(ActiveIndexedComponent):
    _ComponentDataClass = SequenceVarData

    def __new__(cls, *args, **kwds):
        if cls != SequenceVar:
            return super(SequenceVar, cls).__new__(cls)
        if args == ():
            return ScalarSequenceVar.__new__(ScalarSequenceVar)
        else:
            return IndexedSequenceVar.__new__(IndexedSequenceVar)

    def __init__(self, *args, **kwargs):
        self._init_rule = Initializer(kwargs.pop('rule', None))
        self._init_expr = kwargs.pop('expr', None)
        kwargs.setdefault('ctype', SequenceVar)
        super(SequenceVar, self).__init__(*args, **kwargs)

        if self._init_expr is not None and self._init_rule is not None:
            raise ValueError(
                "Cannot specify both rule= and expr= for SequenceVar %s" % (self.name,)
            )

    def _getitem_when_not_present(self, index):
        if index is None and not self.is_indexed():
            obj = self._data[index] = self
        else:
            obj = self._data[index] = self._ComponentDataClass(component=self)
        parent = self.parent_block()
        obj._index = index

        if self._init_rule is not None:
            obj.set_value(self._init_rule(parent, index))
        if self._init_expr is not None:
            obj.set_value(self._init_expr)

        return obj

    def construct(self, data=None):
        """
        Construct the SequenceVarData objects for this SequenceVar
        """
        if self._constructed:
            return
        self._constructed = True

        if is_debug_set(logger):
            logger.debug("Constructing SequenceVar %s" % self.name)

        # Initialize index in case we hit the exception below
        index = None
        try:
            if not self.is_indexed():
                self._getitem_when_not_present(None)
            if self._init_rule is not None:
                for index in self.index_set():
                    self._getitem_when_not_present(index)
        except Exception:
            err = sys.exc_info()[1]
            logger.error(
                "Rule failed when initializing sequence variable for "
                "SequenceVar %s with index %s:\n%s: %s"
                % (self.name, str(index), type(err).__name__, err)
            )
            raise

    def _pprint(self):
        """Print component information."""
        headers = [
            ("Size", len(self)),
            ("Index", self._index_set if self.is_indexed() else None),
        ]
        return (
            headers,
            self.items,
            ("IntervalVars",),
            lambda k, v: ['[' + ', '.join(iv.name for iv in v.interval_vars) + ']'],
        )


class ScalarSequenceVar(SequenceVarData, SequenceVar):
    def __init__(self, *args, **kwds):
        SequenceVarData.__init__(self, component=self)
        SequenceVar.__init__(self, *args, **kwds)
        self._index = UnindexedComponent_index


class IndexedSequenceVar(SequenceVar):
    pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""A module for "flattening" the components in a block-hierarchical model
with respect to common indexing sets

"""

from pyomo.core.base import Block, Reference
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.base.block import SubclassOf
from pyomo.core.base.set import SetProduct
from pyomo.core.base.indexed_component import UnindexedComponent_set, normalize_index
from pyomo.core.base.component import ActiveComponent
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from collections import OrderedDict


def get_slice_for_set(s):
    """
    Get the slice of the proper dimension for a set.
    """
    if s.dimen != 0:
        if not normalize_index.flatten:
            return slice(None)
        else:
            if s.dimen is not None:
                # We will arrive here and fail for sets of dimension
                # UnknownSetDimen.
                return (slice(None),) * s.dimen
            else:
                return (Ellipsis,)
    else:
        # Case for e.g. UnindexedComponent_set
        # Should this be None or tuple()? RBP 202110
        return None


class _NotAnIndex:
    """
    `None` is a valid index, so we use a dummy class to
    denote a slot that needs to get filled with indices
    from our product.
    """

    pass


def _fill_indices(filled_index, index):
    """
    `filled_index` is a list with some entries `_NotAnIndex`.
    We fill those entries with values from `index`, a tuple.
    """
    j = 0
    for i, val in enumerate(filled_index):
        if val is _NotAnIndex:
            filled_index[i] = index[j]
            # `index` is always a tuple, so this is valid
            j += 1
    # Make sure `filled_index` had the same number of vacancies
    # as the original SetProduct has factors. Not _strictly_ necessary.
    assert j == len(index)
    filled_index = tuple(filled_index)

    if len(filled_index) == 1:
        return filled_index[0]
    else:
        return filled_index


def _fill_indices_from_product(partial_index_list, product):
    """
    `partial_index_list` is a list of indices, each corresponding to a
    set. If an entry in `partial_index_list` is `_NotAnIndex`, that
    slot will get filled in by an entry from `product`.

    `product` is a `SetProduct` with as many "factors" as there are
    missing indices in `partial_index_list`.
    """
    # We will manipulate `normalize_index.flatten`.
    # Store its original value so we can reset it when we're done.
    _normalize_index_flatten = normalize_index.flatten
    try:
        normalize_index.flatten = False
        for index in product:
            # Since `normalize_index.flatten` is False, `index` is a
            # scalar or (tuple of (scalars or tuples)). Conveniently,
            # each entry in the tuple belongs to a single factor set.
            # I do not have to worry about having a product-of-products
            # here because I created the product from "unfactorable sets"

            # We need to generate a new index for every entry of `product`,
            # and want to reuse `partial_index_list` as a starting point,
            # so we copy it here.
            filled_index = list(partial_index_list)

            normalize_index.flatten = _normalize_index_flatten

            # filled_index can now be used in the user's intended way
            # This determines how we will access the component's data
            # with our new index, which is currently _completely_ unflattened
            # (i.e. a tuple-of-tuples, no further nesting).
            #
            # This will not work well if the user's component is indexed by
            # a nested product of sets AND normalize_index.flatten is False.
            # In this case we may try to access a component's data with
            # >>> comp[(1,'a',1)]
            # (each coordinate belongs to its own set) when it expects
            # >>> comp[((1,'a'),1)]
            # because `comp` was created with two set arguments, the first
            # of which was already a product.

            yield (index, _fill_indices(filled_index, index))

            normalize_index.flatten = False
            # Want to get the unflattened factors when we advance the
            # iterator of `product`
    finally:
        # Reset `normalize_index.flatten`
        normalize_index.flatten = _normalize_index_flatten


def slice_component_along_sets(component, sets, context_slice=None, normalize=None):
    """This function generates all possible slices of the provided component
    along the provided sets. That is, it will iterate over the component's
    other indexing sets and, for each index, yield a slice along the
    sets specified in the call signature.

    Arguments
    ---------
    component: Component
        The component whose slices will be yielded
    sets: ComponentSet
        ComponentSet of Pyomo sets that will be sliced along
    context_slice: IndexedComponent_slice
        If provided, instead of creating a new slice, we will extend this
        one with appropriate getattr and getitem calls.
    normalize: Bool
        If False, the returned index (from the product of "other sets")
        is not normalized, regardless of the value of normalize_index.flatten.
        This is necessary to use this index with _fill_indices.

    Yields
    ------
    tuple
        The first entry is the index in the product of "other sets"
        corresponding to the slice, and the second entry is the slice
        at that index.

    """
    set_set = ComponentSet(sets)
    subsets = list(component.index_set().subsets())
    temp_idx = [get_slice_for_set(s) if s in set_set else _NotAnIndex for s in subsets]
    other_sets = [s for s in subsets if s not in set_set]

    if context_slice is None:
        base_component = component
    else:
        base_component = getattr(context_slice, component.local_name)

    if component.is_indexed():
        # We need to iterate over sets that aren't sliced
        # `c.is_indexed()` covers the case when UnindexedComponent_set
        # is in `other_sets`.
        if other_sets:
            cross_prod = other_sets[0].cross(*other_sets[1:])
        else:
            # If we are only indexed by sets we need to slice, we
            # should just use tuple(temp_idx) as our index. We spoof
            # a cross_prod here so we don't have to repeat the try/except
            # logic below in a separate branch. An empty tuple is the right
            # singleton to work in the embedded call to _fill_indices.
            cross_prod = [tuple()]

        for prod_index, new_index in _fill_indices_from_product(temp_idx, cross_prod):
            try:
                if normalize_index.flatten:
                    # This index is always normalized if normalize_index.flatten
                    # is True. I have not encountered a situation where
                    # "denormalization" makes sense here.
                    # As normalization is also done in the IndexedComponent,
                    # normalizing here primarily just affects what the resulting
                    # slice "looks like." E.g. slice(None) vs (slice(None),).
                    # This has implications for generating CUIDs from these
                    # slices, where we would like consistency in the string
                    # representation.
                    # TODO: Should CUID normalize (slice(None),)?
                    new_index = normalize_index(new_index)
                c_slice = base_component[new_index]
                if type(c_slice) is IndexedComponent_slice:
                    # This is just to make sure we do not have an
                    # empty slice.
                    #
                    # Note that c_slice is not necessarily a slice.
                    # We enter this loop even if no sets need slicing.
                    try:
                        next(iter(c_slice.duplicate()))
                    except IndexError:
                        if normalize_index.flatten:
                            raise
                        # There is an edge case where when we are not
                        # flattening indices the dimensionality of an
                        # index can change between a SetProduct and the
                        # member Sets: the member set can have dimen>1
                        # (or even None!), but the dimen of that portion
                        # of the SetProduct is always 1.  Since we are
                        # just checking that the c_slice isn't
                        # completely empty, we will allow matching with
                        # an Ellipsis
                        _empty = True
                        try:
                            next(iter(base_component[...]))
                            _empty = False
                        except:
                            pass
                        if _empty:
                            raise
                if (normalize is None and normalize_index.flatten) or normalize:
                    # Most users probably want this index to be normalized,
                    # so they can more conveniently use it as a key in a
                    # mapping. (E.g. they will get "a" as opposed to ("a",).)
                    # However, to use it in the calling routine
                    # generate_sliced_components, we need this index to not
                    # have been normalized, so that indices are tuples,
                    # partitioned according to their "factor sets."
                    # This is why we allow the argument normalize=False to
                    # override normalize_index.flatten.
                    prod_index = normalize_index(prod_index)
                yield prod_index, c_slice
            except StopIteration:
                # We have an empty slice for some reason, e.g.
                # a coordinate of `new_index` from the cross
                # product was skipped in the original component.
                pass
            except KeyError:
                # We are creating scalar components from a product of
                # sets. Components may be undefined for certain indices.
                # We want to simply skip that index and move on.
                pass
    else:
        # Component is a data object
        c_slice = base_component
        yield (), c_slice


def generate_sliced_components(
    b, index_stack, slice_, sets, ctype, index_map, active=None
):
    """Recursively generate slices of the specified ctype along the
    specified sets

    Parameters
    ----------

    b: BlockData
        Block whose components will be sliced

    index_stack: list
        Sets above ``b`` in the block hierarchy, including on its parent
        component, that have been sliced. This is necessary to return the
        sets that have been sliced.

    slice_: IndexedComponent_slice or BlockData
        Slice generated so far.  This function will yield extensions to
        this slice at the current level of the block hierarchy.

    sets: ComponentSet of Pyomo sets
        Sets that will be sliced

    ctype: Subclass of Component
        Type of components to generate

    index_map: ComponentMap
        Map from (some of) the specified sets to a "representative index"
        to use when descending into subblocks. While this map does not need
        to contain every set in the sliced sets, it must not contain any
        sets that will not be sliced.

    active: Bool or None
        If not None, this is a boolean flag used to filter component objects
        by their active status.

    Yields
    ------

    Tuple of Sets and an IndexedComponent_slice or ComponentData
        The sets indexing the returned component or slice. If the component
        is indexed, an IndexedComponent_slice is returned. Otherwise, a
        ComponentData is returned.

    """
    if type(slice_) is IndexedComponent_slice:
        context_slice = slice_.duplicate()
    else:
        context_slice = None

    # If active argument is specified and does not match the block's
    # active flag, we return immediately. This matches the behavior of
    # component_objects. We only need this check as we may modify the
    # active argument sent to component_objects if ctype is not an
    # ActiveComponent type.
    if active is not None and active != b.active:
        return

    # Define this class so we don't have to call issubclass again later.
    check_active = issubclass(ctype, ActiveComponent) and (active != None)

    # If active=False and ctype is not an ActiveComponent (e.g. it is Var)
    # we will not generate any components. To prevent this, only pass the
    # active argument if we are looking for active components.
    c_active = active if check_active else None

    # Looks for components indexed by specified sets immediately in our block.
    for c in b.component_objects(ctype, descend_into=False, active=c_active):
        subsets = list(c.index_set().subsets())
        new_sets = [s for s in subsets if s in sets]
        # Extend our "index stack"
        sliced_sets = index_stack + new_sets

        # Extend our slice with this component
        for idx, new_slice in slice_component_along_sets(
            c, sets, context_slice=context_slice, normalize=False
        ):
            # If we have to check activity, check data objects defined by
            # slice. If any match, we yield the slice. This is done for
            # compatibility with the behavior when slicing blocks, where
            # we can only descend into a block that matches our active flag.
            #
            # Note that new_slice can be a data object. This happens if the
            # component doesn't contain any sets we are slicing, i.e. new_sets
            # is empty.
            if (
                # Yield if (a) we're not checking activity
                not check_active
                # or (b) we have not sliced and data object activity matches
                or (not sliced_sets and new_slice.active == c_active)
                # or (c) we did slice and *any* data object activity matches
                or (
                    sliced_sets
                    and any(data.active == c_active for data in new_slice.duplicate())
                )
            ):
                yield sliced_sets, new_slice

    # We now descend into subblocks
    for sub in b.component_objects(Block, descend_into=False, active=active):
        subsets = list(sub.index_set().subsets())
        new_sets = [s for s in subsets if s in sets]

        # Extend stack with new matched indices.
        index_stack.extend(new_sets)

        # Need to construct an index to descend into for each slice-of-block
        # we are about generate.
        # Note that any remaining _NotAnIndex placeholders after this loop
        # will be replaced with the corresponding indices of the non-sliced
        # sets.
        given_descend_idx = [_NotAnIndex for _ in subsets]
        for i, s in enumerate(subsets):
            # NOTE: index_map better only contain sets that we are slicing.
            if s in index_map:
                # Use a user-given index if available.
                given_descend_idx[i] = index_map[s]
                if s not in sets:
                    raise RuntimeError(
                        "Encountered a specified index for a set %s that we"
                        " are not slicing. This is not supported" % s
                    )
            elif s in sets:
                # Otherwise use a slice. We will advanced the slice iter
                # to try to get a concrete component from this slice.
                given_descend_idx[i] = get_slice_for_set(s)

        # Generate slices from this sub-block
        for idx, new_slice in slice_component_along_sets(
            sub, sets, context_slice=context_slice, normalize=False
        ):
            # TODO: Can this branch happen outside of the loop?
            # If it's not indexed, we don't need to slice...
            if sub.is_indexed():
                # fill any remaining placeholders with the "index" of our slice
                descend_idx = _fill_indices(list(given_descend_idx), idx)
                # create a slice-or-data object
                descend_data = sub[descend_idx]
                if type(descend_data) is IndexedComponent_slice:
                    try:
                        slice_iter = iter(descend_data)
                        # Try to find a data object defined by the slice
                        # that matches the active argument. In doing so,
                        # we treat a slice as inactive if all of its data
                        # objects are inactive. We need to find a data obj
                        # with the correct active flag, otherwise we run into
                        # problems when we descend (component_objects will
                        # not yield anything).
                        _data = next(slice_iter)
                        while active is not None and _data.active != active:
                            _data = next(slice_iter)
                        descend_data = _data
                    except StopIteration:
                        # For this particular idx, we have no BlockData
                        # to descend into.
                        continue
                elif active is not None and descend_data.active != active:
                    # descend_data is a BlockData object. This particular
                    # BlockData was specified by the index map. In this case,
                    # we want to respect "activity".
                    continue
            else:
                # Have encountered a ScalarBlock. Do not need to check the
                # active flag as this came straight from component_objects.
                descend_data = sub

            # Recursively generate sliced components from this data object
            for st, v in generate_sliced_components(
                descend_data,
                index_stack,
                new_slice,
                sets,
                ctype,
                index_map,
                active=active,
            ):
                yield tuple(st), v

        # pop the index sets of the block whose sub-components
        # we just finished iterating over.
        for _ in new_sets:
            index_stack.pop()


def flatten_components_along_sets(m, sets, ctype, indices=None, active=None):
    """This function iterates over components (recursively) contained
    in a block and partitions their data objects into components
    indexed only by the specified sets.

    Parameters
    ----------

    m: BlockData
        Block whose components (and their sub-components) will be
        partitioned

    sets: Tuple of Pyomo Sets
        Sets to be sliced. Returned components will be indexed by
        some combination of these sets, if at all.

    ctype: Subclass of Component
        Type of component to identify and partition

    indices: Iterable or ComponentMap
        Indices of sets to use when descending into subblocks. If an
        iterable is provided, the order corresponds to the order in
        ``sets``. If a ``ComponentMap`` is provided, the keys must be
        in ``sets``.

    active: Bool or None
        If not None, this is a boolean flag used to filter component objects
        by their active status. A reference-to-slice is returned if any data
        object defined by the slice matches this flag.

    Returns
    -------

    List of tuples of Sets, list of lists of Components
        The first entry is a list of tuples of Pyomo Sets. The second is a
        list of lists of Components, indexed by the corresponding sets in
        the first list. If the components are unindexed, ComponentData are
        returned and the tuple of sets contains only UnindexedComponent_set.
        If the components are indexed, they are references-to-slices.

    """
    set_of_sets = ComponentSet(sets)
    if indices is None:
        index_map = ComponentMap()
    elif type(indices) is ComponentMap:
        index_map = indices
    else:
        index_map = ComponentMap(zip(sets, indices))
    for s, idx in index_map.items():
        if idx not in s:
            raise ValueError(
                "%s is a bad index for set %s. \nPlease provide an index "
                "that is in the set." % (idx, s.name)
            )
        if s not in set_of_sets:
            raise RuntimeError(
                "Index specified for set %s that is not one of the sets"
                " that will be sliced. Indices should only be provided"
                " for sets that will be sliced." % s.name
            )
    index_stack = []

    # Using these two `OrderedDict`s is a workaround because I can't
    # reliably use tuples of components as keys in a `ComponentMap`.
    sets_dict = OrderedDict()
    comps_dict = OrderedDict()
    for index_sets, slice_ in generate_sliced_components(
        m, index_stack, m, set_of_sets, ctype, index_map, active=active
    ):
        # Note that index_sets should always be a tuple, never a scalar.

        # TODO: Potentially re-order sets at this point.
        # In this way (time, space) would have the same key as (space, time).
        # They we'd have to somehow "swap indexing sets" when we create
        # the reference below.
        key = tuple(id(c) for c in index_sets)
        if key not in sets_dict:
            if len(key) == 0:
                sets_dict[key] = (UnindexedComponent_set,)
            else:
                sets_dict[key] = index_sets
        if key not in comps_dict:
            comps_dict[key] = []
        if len(key) == 0:
            comps_dict[key].append(slice_)
        else:
            # If the user wants to change these flags, they can access the
            # slice via the `referent` attribute of each reference component.
            slice_.attribute_errors_generate_exceptions = False
            slice_.key_errors_generate_exceptions = False
            comps_dict[key].append(Reference(slice_))

    # list-of-tuples of Sets:
    sets_list = list(sets for sets in sets_dict.values())
    # list-of-lists of components:
    comps_list = list(comps for comps in comps_dict.values())
    # E.g. we return: (
    #          [(time, space), (time,)],
    #          [[some_component, ...], [other, ...]],
    #      )                            ^ These components are indexed by time
    #            ^ These components are indexed by time and space
    return sets_list, comps_list


def flatten_dae_components(model, time, ctype, indices=None, active=None):
    """Partitions components into ComponentData and Components indexed only
    by the provided set.

    Parameters
    ----------

    model: BlockData
        Block whose components are partitioned

    time: Set
        Indexing by this set (and only this set) will be preserved in the
        returned components.

    ctype: Subclass of Component
        Type of component to identify, partition, and return

    indices: Tuple or ComponentMap
        Contains the index of the specified set to be used when descending
        into blocks

    active: Bool or None
        If provided, used as a filter to only return components with the
        specified active flag. A reference-to-slice is returned if any
        data object defined by the slice matches this flag.

    Returns
    -------
    List of ComponentData, list of Component
        The first list contains ComponentData for all components not
        indexed by the provided set. The second contains references-to
        -slices for all components indexed by the provided set.

    """
    target = ComponentSet((time,))
    sets_list, comps_list = flatten_components_along_sets(
        model, target, ctype, indices=indices, active=active
    )
    # Initialize these variables as, if no components of either category are
    # found, we expect to get an empty list.
    scalar_comps = []
    dae_comps = []
    for sets, comps in zip(sets_list, comps_list):
        if len(sets) == 1 and sets[0] is time:
            dae_comps = comps
        elif len(sets) == 0 or (len(sets) == 1 and sets[0] is UnindexedComponent_set):
            scalar_comps = comps
        else:
            raise RuntimeError(
                "Invalid model for `flatten_dae_components`.\n"
                "This can happen if your model has components that are\n"
                "indexed by time (explicitly or implicitly) multiple times."
            )
    return scalar_comps, dae_comps



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests for the GDPopt solver plugin."""

from contextlib import redirect_stdout
from io import StringIO
import logging
from math import fabs
from os.path import join, normpath

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept
from pyomo.common.collections import Bunch
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.common.fileutils import import_file, PYOMO_ROOT_DIR
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    add_util_block,
    add_disjunct_list,
    add_constraints_by_disjunct,
    add_global_constraint_list,
)
import pyomo.contrib.gdpopt.tests.common_tests as ct
from pyomo.contrib.gdpopt.util import is_feasible, time_code
from pyomo.contrib.mcpp.pyomo_mcpp import mcpp_available
from pyomo.contrib.gdpopt.solve_discrete_problem import (
    solve_MILP_discrete_problem,
    distinguish_mip_infeasible_or_unbounded,
)
from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    Integers,
    LogicalConstraint,
    maximize,
    Objective,
    RangeSet,
    TransformationFactory,
    SolverFactory,
    sqrt,
    value,
    Var,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.gdp.tests import models
from pyomo.opt import TerminationCondition

exdir = normpath(join(PYOMO_ROOT_DIR, 'examples', 'gdp'))

mip_solver = 'glpk'
nlp_solver = 'ipopt'
global_nlp_solver = 'baron'
global_nlp_solver_args = dict()
minlp_solver = 'baron'
LOA_solvers = (mip_solver, nlp_solver)
GLOA_solvers = (mip_solver, global_nlp_solver, minlp_solver)
LOA_solvers_available = all(SolverFactory(s).available() for s in LOA_solvers)
GLOA_solvers_available = all(SolverFactory(s).available() for s in GLOA_solvers)
license_available = (
    SolverFactory(global_nlp_solver).license_is_valid()
    if GLOA_solvers_available
    else False
)

gurobi_available = (
    SolverFactory('gurobi').available(exception_flag=False)
    and SolverFactory('gurobi').license_is_valid()
)


class TestGDPoptUnit(unittest.TestCase):
    """Real unit tests for GDPopt"""

    @unittest.skipUnless(
        SolverFactory(mip_solver).available(), "MIP solver not available"
    )
    def test_solve_discrete_problem_unbounded(self):
        m = ConcreteModel()
        m.GDPopt_utils = Block()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.z = Var()
        # Include a disjunction so that we don't default to just a MIP solver
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=m.z)
        m.GDPopt_utils.variable_list = [m.x, m.y, m.z]
        m.GDPopt_utils.disjunct_list = [m.d._autodisjuncts[0], m.d._autodisjuncts[1]]
        TransformationFactory('gdp.bigm').apply_to(m)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.WARNING):
            solver = SolverFactory('gdpopt.loa')
            dummy = Block()
            dummy.timing = Bunch()
            with time_code(dummy.timing, 'main', is_main_timer=True):
                tc = solve_MILP_discrete_problem(
                    m.GDPopt_utils, dummy, solver.CONFIG(dict(mip_solver=mip_solver))
                )
            self.assertIn(
                "Discrete problem was unbounded. Re-solving with "
                "arbitrary bound values",
                output.getvalue().strip(),
            )
        self.assertIs(tc, TerminationCondition.unbounded)

    @unittest.skipUnless(
        SolverFactory(mip_solver).available(), "MIP solver not available"
    )
    def test_solve_lp(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-5, 5))
        m.c = Constraint(expr=m.x >= 1)
        m.o = Objective(expr=m.x)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            results = SolverFactory('gdpopt.loa').solve(m, mip_solver=mip_solver)
            self.assertIn(
                "Your model is an LP (linear program).", output.getvalue().strip()
            )
            self.assertAlmostEqual(value(m.o.expr), 1)

            self.assertEqual(results.problem.number_of_binary_variables, 0)
            self.assertEqual(results.problem.number_of_integer_variables, 0)
            self.assertEqual(results.problem.number_of_disjunctions, 0)
            self.assertAlmostEqual(results.problem.lower_bound, 1)
            self.assertAlmostEqual(results.problem.upper_bound, 1)

    @unittest.skipUnless(gurobi_available, 'Gurobi not available')
    def test_solve_nlp(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-5, 5))
        m.c = Constraint(expr=m.x >= 1)
        m.o = Objective(expr=m.x**2)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            results = SolverFactory('gdpopt.loa').solve(m, nlp_solver='gurobi')
            self.assertIn(
                "Your model is an NLP (nonlinear program).", output.getvalue().strip()
            )
            self.assertAlmostEqual(value(m.o.expr), 1)

            self.assertEqual(results.problem.number_of_binary_variables, 0)
            self.assertEqual(results.problem.number_of_integer_variables, 0)
            self.assertEqual(results.problem.number_of_disjunctions, 0)
            self.assertAlmostEqual(results.problem.lower_bound, 1)
            self.assertAlmostEqual(results.problem.upper_bound, 1)

    @unittest.skipUnless(
        SolverFactory(mip_solver).available(), "MIP solver not available"
    )
    def test_solve_constant_obj(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-5, 5))
        m.c = Constraint(expr=m.x >= 1)
        m.o = Objective(expr=1)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            SolverFactory('gdpopt.loa').solve(m, mip_solver=mip_solver)
            self.assertIn(
                "Your model is an LP (linear program).", output.getvalue().strip()
            )
            self.assertAlmostEqual(value(m.o.expr), 1)

    @unittest.skipUnless(
        SolverFactory(nlp_solver).available(), 'NLP solver not available'
    )
    def test_no_objective(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-5, 5))
        m.c = Constraint(expr=m.x**2 >= 1)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.WARNING):
            SolverFactory('gdpopt.loa').solve(m, nlp_solver=nlp_solver)
            self.assertIn(
                "Model has no active objectives. Adding dummy objective.",
                output.getvalue().strip(),
            )

        # check that the dummy objective is removed after the solve (else
        # repeated solves result in the error about multiple active objectives
        # on the model)
        self.assertIsNone(m.component("dummy_obj"))

    def test_multiple_objectives(self):
        m = ConcreteModel()
        m.x = Var()
        m.o = Objective(expr=m.x)
        m.o2 = Objective(expr=m.x + 1)
        with self.assertRaisesRegex(ValueError, "Model has multiple active objectives"):
            SolverFactory('gdpopt.loa').solve(m)

    def test_is_feasible_function(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 3), initialize=2)
        m.c = Constraint(expr=m.x == 2)
        GDP_LOA_Solver = SolverFactory('gdpopt.loa')
        self.assertTrue(is_feasible(m, GDP_LOA_Solver.CONFIG()))

        m.c2 = Constraint(expr=m.x <= 1)
        self.assertFalse(is_feasible(m, GDP_LOA_Solver.CONFIG()))

        m = ConcreteModel()
        m.x = Var(bounds=(0, 3), initialize=2)
        m.c = Constraint(expr=m.x >= 5)
        self.assertFalse(is_feasible(m, GDP_LOA_Solver.CONFIG()))

        m = ConcreteModel()
        m.x = Var(bounds=(3, 3), initialize=2)
        self.assertFalse(is_feasible(m, GDP_LOA_Solver.CONFIG()))

        m = ConcreteModel()
        m.x = Var(bounds=(0, 1), initialize=2)
        self.assertFalse(is_feasible(m, GDP_LOA_Solver.CONFIG()))

        m = ConcreteModel()
        m.x = Var(bounds=(0, 1), initialize=2)
        m.d = Disjunct()
        with self.assertRaisesRegex(NotImplementedError, "Found active disjunct"):
            is_feasible(m, GDP_LOA_Solver.CONFIG())

    @unittest.skipUnless(gurobi_available, 'Gurobi not available')
    def test_infeasible_or_unbounded_mip_termination(self):
        m = ConcreteModel()
        m.x = Var()
        m.c1 = Constraint(expr=m.x >= 2)
        m.c2 = Constraint(expr=m.x <= 1.9)
        m.obj = Objective(expr=m.x)

        results = SolverFactory('gurobi').solve(m)
        # Gurobi shrugs:
        self.assertEqual(
            results.solver.termination_condition,
            TerminationCondition.infeasibleOrUnbounded,
        )
        # just pretend on the config block--we only need these two:
        config = ConfigDict()
        config.declare('mip_solver', ConfigValue('gurobi'))
        config.declare('mip_solver_args', ConfigValue({}))

        # We tell Gurobi to figure it out
        results, termination_condition = distinguish_mip_infeasible_or_unbounded(
            m, config
        )

        # It's infeasible:
        self.assertEqual(termination_condition, TerminationCondition.infeasible)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )

    def get_GDP_on_block(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-5, 5))
        m.y = Var(bounds=(-2, 6))
        m.b = Block()
        m.b.disjunction = Disjunction(
            expr=[
                [m.x + m.y <= 1, m.y >= 0.5],
                [m.x == 2, m.y == 4],
                [m.x**2 - m.y <= 3],
            ]
        )
        m.disjunction = Disjunction(
            expr=[
                [m.x - m.y <= -2, m.y >= -1],
                [m.x == 0, m.y >= 0],
                [m.y**2 + m.x <= 3],
            ]
        )
        return m

    def test_gloa_cut_generation_ignores_deactivated_constraints(self):
        m = self.get_GDP_on_block()
        m.b.disjunction.disjuncts[0].indicator_var.fix(True)
        m.b.disjunction.disjuncts[1].indicator_var.fix(False)
        m.b.disjunction.disjuncts[2].indicator_var.fix(False)
        m.b.deactivate()
        m.disjunction.disjuncts[0].indicator_var.fix(False)
        m.disjunction.disjuncts[1].indicator_var.fix(True)
        m.disjunction.disjuncts[2].indicator_var.fix(False)

        # set up the util block
        add_util_block(m)
        util_block = m._gdpopt_cuts
        add_disjunct_list(util_block)
        add_constraints_by_disjunct(util_block)
        add_global_constraint_list(util_block)

        TransformationFactory('gdp.bigm').apply_to(m)

        config = ConfigDict()
        config.declare('integer_tolerance', ConfigValue(1e-6))

        gloa = SolverFactory('gdpopt.gloa')

        constraints = list(
            gloa._get_active_untransformed_constraints(util_block, config)
        )
        self.assertEqual(len(constraints), 2)
        c1 = constraints[0]
        c2 = constraints[1]
        self.assertIs(c1.body, m.x)
        self.assertEqual(c1.lower, 0)
        self.assertEqual(c1.upper, 0)
        self.assertIs(c2.body, m.y)
        self.assertEqual(c2.lower, 0)
        self.assertIsNone(c2.upper)

    def test_complain_when_no_algorithm_specified(self):
        m = self.get_GDP_on_block()
        with self.assertRaisesRegex(
            ValueError,
            "No algorithm was specified to the solve method. "
            "Please specify an algorithm or use an "
            "algorithm-specific solver.",
        ):
            SolverFactory('gdpopt').solve(m)

    @unittest.skipIf(
        not LOA_solvers_available,
        "Required subsolvers %s are not available" % (LOA_solvers,),
    )
    def test_solve_block(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.x = Var(bounds=(-5, 5))
        m.b.y = Var(bounds=(-2, 6))
        m.b.disjunction = Disjunction(
            expr=[
                [m.b.x + m.b.y <= 1, m.b.y >= 0.5],
                [m.b.x == 2, m.b.y == 4],
                [m.b.x**2 - m.b.y <= 3],
            ]
        )
        m.disjunction = Disjunction(
            expr=[
                [m.b.x - m.b.y <= -2, m.b.y >= -1],
                [m.b.x == 0, m.b.y >= 0],
                [m.b.y**2 + m.b.x <= 3],
            ]
        )
        m.b.obj = Objective(expr=m.b.x)

        SolverFactory('gdpopt.ric').solve(
            m.b, mip_solver=mip_solver, nlp_solver=nlp_solver
        )

        # There are multiple optimal solutions, so just leave it at this:
        self.assertAlmostEqual(value(m.b.x), -5)
        # We didn't declare any Block on m--it's still just b.
        self.assertEqual(len(m.component_map(Block)), 1)


@unittest.skipIf(
    not LOA_solvers_available,
    "Required subsolvers %s are not available" % (LOA_solvers,),
)
class TestGDPopt(unittest.TestCase):
    """Tests for the GDPopt solver plugin."""

    def test_infeasible_GDP(self):
        """Test for infeasible GDP."""
        m = models.make_infeasible_gdp_model()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.WARNING):
            results = SolverFactory('gdpopt.loa').solve(
                m, mip_solver=mip_solver, nlp_solver=nlp_solver
            )
            self.assertIn(
                "Set covering problem is infeasible.", output.getvalue().strip()
            )

        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.d.disjuncts[0].indicator_var.value)
        self.assertIsNone(m.d.disjuncts[1].indicator_var.value)

        # Test maximization problem infeasibility also
        m.o.sense = maximize
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            results = SolverFactory('gdpopt').solve(
                m,
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
                init_algorithm='no_init',
                algorithm='LOA',
            )
            self.assertIn(
                "GDPopt exiting--problem is infeasible.", output.getvalue().strip()
            )

        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertIsNotNone(results.solver.user_time)
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.d.disjuncts[0].indicator_var.value)
        self.assertIsNone(m.d.disjuncts[1].indicator_var.value)

    def test_infeasible_gdp_max_binary(self):
        """Test that max binary initialization catches infeasible GDP too"""
        m = models.make_infeasible_gdp_model()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.DEBUG):
            results = SolverFactory('gdpopt.loa').solve(
                m,
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
                init_algorithm='max_binary',
            )
            self.assertIn(
                "MILP relaxation for initialization was infeasible. "
                "Problem is infeasible.",
                output.getvalue().strip(),
            )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )

    def test_unbounded_gdp_minimization(self):
        m = ConcreteModel()
        m.GDPopt_utils = Block()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.z = Var()
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=m.z)
        m.GDPopt_utils.variable_list = [m.x, m.y, m.z]
        m.GDPopt_utils.disjunct_list = [m.d._autodisjuncts[0], m.d._autodisjuncts[1]]
        results = SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.unbounded
        )

    def test_unbounded_gdp_maximization(self):
        m = ConcreteModel()
        m.GDPopt_utils = Block()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.z = Var()
        m.d = Disjunction(expr=[[m.x + m.y <= 5], [m.x - m.y >= 3]])
        m.o = Objective(expr=m.z, sense=maximize)
        m.GDPopt_utils.variable_list = [m.x, m.y, m.z]
        m.GDPopt_utils.disjunct_list = [m.d._autodisjuncts[0], m.d._autodisjuncts[1]]
        results = SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.unbounded
        )

    # [ESJ 5/16/22]: Using Gurobi for this test because glpk seems to get angry
    # on Windows when the MIP is arbitrarily bounded with the large bounds. And
    # I think I blame glpk...
    @unittest.skipUnless(gurobi_available, "Gurobi solver not available")
    def test_GDP_nonlinear_objective(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=m.x**2)
        SolverFactory('gdpopt.loa').solve(m, mip_solver='gurobi', nlp_solver=nlp_solver)
        self.assertAlmostEqual(value(m.o), 0)

        m = ConcreteModel()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=-m.x**2, sense=maximize)
        print("second")
        SolverFactory('gdpopt.loa').solve(m, mip_solver='gurobi', nlp_solver=nlp_solver)
        self.assertAlmostEqual(value(m.o), 0)

    def test_nested_disjunctions_set_covering(self):
        # This test triggers the InfeasibleConstraintException in
        # deactivate_trivial_constraints in one of the subproblem solves during
        # initialization. This makes sure we get the correct answer anyway, as
        # there is a feasible solution.
        m = models.makeNestedNonlinearModel()
        SolverFactory('gdpopt.loa').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            init_algorithm='set_covering',
        )
        self.assertAlmostEqual(value(m.x), sqrt(2) / 2)
        self.assertAlmostEqual(value(m.y), sqrt(2) / 2)
        self.assertTrue(value(m.disj.disjuncts[1].indicator_var))
        self.assertFalse(value(m.disj.disjuncts[0].indicator_var))
        self.assertTrue(value(m.d1.indicator_var))
        self.assertFalse(value(m.d2.indicator_var))

    def test_equality_propagation_infeasibility_in_subproblems(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-10, 10))
        m.y = Var(bounds=(-10, 10))
        m.disj = Disjunction(
            expr=[[m.x == m.y, m.y == 2], [m.y == 8], [m.x + m.y >= 4, m.y == m.x + 1]]
        )
        m.cons = Constraint(expr=m.x == 3)
        m.obj = Objective(expr=m.x + m.y)
        SolverFactory('gdpopt').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            init_algorithm='set_covering',
            algorithm='RIC',
        )
        self.assertAlmostEqual(value(m.x), 3)
        self.assertAlmostEqual(value(m.y), 4)
        self.assertFalse(value(m.disj.disjuncts[0].indicator_var))
        self.assertFalse(value(m.disj.disjuncts[1].indicator_var))
        self.assertTrue(value(m.disj.disjuncts[2].indicator_var))

    def test_bound_infeasibility_in_subproblems(self):
        m = ConcreteModel()
        m.x = Var(bounds=(2, 4))
        m.y = Var(bounds=(5, 10))
        m.disj = Disjunction(expr=[[m.x == m.y, m.x + m.y >= 8], [m.x == 4]])
        m.obj = Objective(expr=m.x + m.y)
        SolverFactory('gdpopt.ric').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            init_algorithm='set_covering',
            tee=True,
        )
        self.assertAlmostEqual(value(m.x), 4)
        self.assertAlmostEqual(value(m.y), 5)
        self.assertFalse(value(m.disj.disjuncts[0].indicator_var))
        self.assertTrue(value(m.disj.disjuncts[1].indicator_var))

    def test_subproblem_preprocessing_encounters_trivial_constraints(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 10))
        m.z = Var(bounds=(-10, 10))
        m.disjunction = Disjunction(expr=[[m.x == 0, m.z >= 4], [m.x + m.z <= 0]])
        m.cons = Constraint(expr=m.x * m.z <= 0)
        m.obj = Objective(expr=-m.z)
        m.disjunction.disjuncts[0].indicator_var.fix(True)
        m.disjunction.disjuncts[1].indicator_var.fix(False)
        SolverFactory('gdpopt.ric').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            init_algorithm='fix_disjuncts',
        )
        # The real test is that this doesn't throw an error when we preprocess
        # to solve the first subproblem (in the initialization). The nonlinear
        # constraint becomes trivial, which we need to make sure is handled
        # correctly.
        self.assertEqual(value(m.x), 0)
        self.assertEqual(value(m.z), 10)
        self.assertTrue(value(m.disjunction.disjuncts[0].indicator_var))
        self.assertFalse(value(m.disjunction.disjuncts[1].indicator_var))

    def make_convex_circle_and_circle_slice_disjunction(self):
        m = ConcreteModel()
        # x will only appear in nonlinear expressions
        m.x = Var(bounds=(-10, 18))
        m.y = Var(bounds=(0, 7))
        m.obj = Objective(expr=m.x**2 + m.y)
        m.disjunction = Disjunction(
            expr=[
                [m.x**2 + m.y**2 <= 3, m.y >= 1],
                (m.x - 3) ** 2 + (m.y - 2) ** 2 <= 1,
            ]
        )

        return m

    def test_some_vars_only_in_subproblem(self):
        # Even variables that only appear in nonlinear constraints will be
        # introduced to the discrete problem via the cuts, so we check that
        # that works out.
        m = self.make_convex_circle_and_circle_slice_disjunction()

        results = SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(results.problem.upper_bound, 1)
        self.assertAlmostEqual(value(m.x), 0)
        self.assertAlmostEqual(value(m.y), 1)

    def test_fixed_vars_honored(self):
        # Make sure that the algorithm doesn't stomp on user-fixed variables
        m = self.make_convex_circle_and_circle_slice_disjunction()

        # first, force ourselves into the suboptimal disjunct
        m.disjunction.disjuncts[0].indicator_var.fix(False)
        SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertTrue(value(m.disjunction.disjuncts[1].indicator_var))
        self.assertFalse(value(m.disjunction.disjuncts[0].indicator_var))
        self.assertTrue(m.disjunction.disjuncts[0].indicator_var.fixed)
        self.assertAlmostEqual(value(m.x), 2.029, places=3)
        self.assertAlmostEqual(value(m.y), 1.761, places=3)
        self.assertAlmostEqual(value(m.obj), 5.878, places=3)

        # Now, do it by fixing a continuous variable
        m.disjunction.disjuncts[0].indicator_var.fixed = False
        m.x.fix(3)
        SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertTrue(value(m.disjunction.disjuncts[1].indicator_var))
        self.assertFalse(value(m.disjunction.disjuncts[0].indicator_var))
        self.assertEqual(value(m.x), 3)
        self.assertTrue(m.x.fixed)
        self.assertAlmostEqual(value(m.y), 1)
        self.assertAlmostEqual(value(m.obj), 10)

    def test_ignore_set_for_oa_cuts(self):
        m = self.make_convex_circle_and_circle_slice_disjunction()

        m.disjunction.disjuncts[1].GDPopt_ignore_OA = [
            m.disjunction.disjuncts[1].constraint[1]
        ]
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.DEBUG):
            SolverFactory('gdpopt.loa').solve(
                m, mip_solver=mip_solver, nlp_solver=nlp_solver
            )
            self.assertIn(
                'OA cut addition for '
                'disjunction_disjuncts[1].constraint[1] skipped '
                'because it is in the ignore set.',
                output.getvalue().strip(),
            )
        # and the solution is optimal
        self.assertAlmostEqual(value(m.x), 0)
        self.assertAlmostEqual(value(m.y), 1)

    def test_reverse_numeric_differentiation_in_LOA(self):
        # Having more than 1000 variables in the same constraint expression will
        # trigger reverse numeric differentiation. We make sure that works with
        # a disjunction between a pair of hyperspheres in 1300-dimensional
        # space.
        m = ConcreteModel()
        m.s = RangeSet(1300)
        m.x = Var(m.s, bounds=(-10, 10))
        m.d1 = Disjunct()
        m.d1.hypersphere = Constraint(expr=sum(m.x[i] ** 2 for i in m.s) <= 1)
        m.d2 = Disjunct()
        m.d2.translated_hyper_sphere = Constraint(
            expr=sum((m.x[i] - i) ** 2 for i in m.s) <= 1
        )
        m.disjunction = Disjunction(expr=[m.d1, m.d2])
        # we'll go in the sphere centered at (0,0,...,0)
        m.obj = Objective(expr=sum(m.x[i] for i in m.s))

        results = SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )

        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertTrue(value(m.d1.indicator_var))
        self.assertFalse(value(m.d2.indicator_var))
        x_val = -sqrt(1300) / 1300
        for x in m.x.values():
            self.assertAlmostEqual(value(x), x_val)
        self.assertAlmostEqual(results.problem.upper_bound, 1300 * x_val, places=6)

    def test_logical_constraints_on_disjuncts(self):
        m = models.makeLogicalConstraintsOnDisjuncts()
        SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertAlmostEqual(value(m.x), 8)

    def test_logical_constraints_on_disjuncts_nonlinear_convex(self):
        m = models.makeLogicalConstraintsOnDisjuncts_NonlinearConvex()
        results = SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(value(m.x), 4)

    def test_nested_disjunctions_no_init(self):
        m = models.makeNestedNonlinearModel()
        SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver, init_algorithm='no_init'
        )
        self.assertAlmostEqual(value(m.x), sqrt(2) / 2)
        self.assertAlmostEqual(value(m.y), sqrt(2) / 2)

    def test_nested_disjunctions_max_binary(self):
        m = models.makeNestedNonlinearModel()
        SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver, init_algorithm='max_binary'
        )
        self.assertAlmostEqual(value(m.x), sqrt(2) / 2)
        self.assertAlmostEqual(value(m.y), sqrt(2) / 2)

    def test_boolean_vars_on_disjuncts(self):
        m = models.makeBooleanVarsOnDisjuncts()
        SolverFactory('gdpopt.loa').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertAlmostEqual(value(m.x), 8)

    def test_LOA_8PP_default_init(self):
        """Test logic-based outer approximation with 8PP."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(
            eight_process, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_iteration_limit(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            results = SolverFactory('gdpopt.loa').solve(
                eight_process, mip_solver=mip_solver, nlp_solver=nlp_solver, iterlim=2
            )
            self.assertIn(
                "GDPopt unable to converge bounds within iteration "
                "limit of 2 iterations.",
                output.getvalue().strip(),
            )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxIterations
        )

    def test_time_limit(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            results = SolverFactory('gdpopt.loa').solve(
                eight_process,
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
                time_limit=1,
            )
            self.assertIn(
                "GDPopt exiting--Did not converge bounds before "
                "time limit of 1 seconds.",
                output.getvalue().strip(),
            )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxTimeLimit
        )

    @unittest.skipUnless(
        license_available, "No BARON license--8PP logical problem exceeds demo size"
    )
    def test_LOA_8PP_logical_default_init(self):
        """Test logic-based outer approximation with 8PP."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_logical.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(
            eight_process, mip_solver=mip_solver, nlp_solver=nlp_solver, tee=False
        )
        ct.check_8PP_logical_solution(self, eight_process, results)

    @unittest.skipUnless(
        SolverFactory('gams').available(exception_flag=False),
        'GAMS solver not available',
    )
    def test_LOA_8PP_gams_solver(self):
        # Make sure that the duals are still correct
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(
            eight_process,
            mip_solver=mip_solver,
            nlp_solver='gams',
            max_slack=0,
            tee=False,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_LOA_8PP_force_NLP(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(
            eight_process,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            force_subproblem_nlp=True,
            tee=False,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_LOA_strip_pack_default_init(self):
        """Test logic-based outer approximation with strip packing."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.loa').solve(
            strip_pack, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    def test_LOA_strip_pack_logical_constraints(self):
        """Test logic-based outer approximation with variation of strip
        packing with some logical constraints."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        # add logical constraints
        strip_pack.Rec3AboveOrBelowRec1 = LogicalConstraint(
            expr=strip_pack.no_overlap[1, 3]
            .disjuncts[2]
            .indicator_var.lor(strip_pack.no_overlap[1, 3].disjuncts[3].indicator_var)
        )
        strip_pack.Rec3RightOrLeftOfRec2 = LogicalConstraint(
            expr=strip_pack.no_overlap[2, 3]
            .disjuncts[0]
            .indicator_var.lor(strip_pack.no_overlap[2, 3].disjuncts[1].indicator_var)
        )
        SolverFactory('gdpopt.loa').solve(
            strip_pack,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            subproblem_presolve=False,
        )  # skip preprocessing for linear problem
        # and so that we test everything is still
        # fine
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 13) <= 1e-2)

    @unittest.pytest.mark.expensive
    def test_LOA_constrained_layout_default_init(self):
        """Test LOA with constrained layout."""
        exfile = import_file(join(exdir, 'constrained_layout', 'cons_layout_model.py'))
        cons_layout = exfile.build_constrained_layout_model()
        SolverFactory('gdpopt.loa').solve(
            cons_layout,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            iterlim=120,
            max_slack=5,  # problem is convex, so can decrease slack
        )
        objective_value = value(cons_layout.min_dist_cost.expr)
        self.assertTrue(
            fabs(objective_value - 41573) <= 200,
            "Objective value of %s instead of 41573" % objective_value,
        )

    def test_LOA_8PP_maxBinary(self):
        """Test logic-based OA with max_binary initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(
            eight_process,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        ct.check_8PP_solution(self, eight_process, results)

    @unittest.skipUnless(
        license_available, "No BARON license--8PP logical problem exceeds demo size"
    )
    def test_LOA_8PP_logical_maxBinary(self):
        """Test logic-based OA with max_binary initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_logical.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(
            eight_process,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        ct.check_8PP_logical_solution(self, eight_process, results)

    def test_LOA_strip_pack_maxBinary(self):
        """Test LOA with strip packing using max_binary initialization."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.loa').solve(
            strip_pack,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    def test_LOA_strip_pack_maxBinary_logical_constraints(self):
        """Test LOA with strip packing using max_binary initialization and
        logical constraints."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        # add logical constraints
        strip_pack.Rec3AboveOrBelowRec1 = LogicalConstraint(
            expr=strip_pack.no_overlap[1, 3]
            .disjuncts[2]
            .indicator_var.lor(strip_pack.no_overlap[1, 3].disjuncts[3].indicator_var)
        )
        strip_pack.Rec3RightOrLeftOfRec2 = LogicalConstraint(
            expr=strip_pack.no_overlap[2, 3]
            .disjuncts[0]
            .indicator_var.lor(strip_pack.no_overlap[2, 3].disjuncts[1].indicator_var)
        )
        SolverFactory('gdpopt.loa').solve(
            strip_pack,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 13) <= 1e-2)

    def test_LOA_8PP_fixed_disjuncts(self):
        """Test LOA with 8PP using fixed disjuncts initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        initialize = [
            # Use units 1, 4, 7, 8
            eight_process.use_unit_1or2.disjuncts[0],
            eight_process.use_unit_3ornot.disjuncts[1],
            eight_process.use_unit_4or5ornot.disjuncts[0],
            eight_process.use_unit_6or7ornot.disjuncts[1],
            eight_process.use_unit_8ornot.disjuncts[0],
        ]
        for disj in eight_process.component_data_objects(Disjunct):
            if disj in initialize:
                disj.binary_indicator_var.set_value(1)
            else:
                disj.binary_indicator_var.set_value(0)
        results = SolverFactory('gdpopt.loa').solve(
            eight_process,
            init_algorithm='fix_disjuncts',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_LOA_custom_disjuncts_with_silly_components_in_list(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()

        eight_process.goofy = Disjunct()
        eight_process.goofy.deactivate()

        initialize = [
            # Use units 1, 4, 7, 8
            [
                eight_process.use_unit_1or2.disjuncts[0],
                eight_process.use_unit_3ornot.disjuncts[1],
                eight_process.use_unit_4or5ornot.disjuncts[0],
                eight_process.use_unit_6or7ornot.disjuncts[1],
                eight_process.use_unit_8ornot.disjuncts[0],
                eight_process.goofy,
            ],
            # Use units 2, 4, 6, 8
            [
                eight_process.use_unit_1or2.disjuncts[1],
                eight_process.use_unit_3ornot.disjuncts[1],
                eight_process.use_unit_4or5ornot.disjuncts[0],
                eight_process.use_unit_6or7ornot.disjuncts[0],
                eight_process.use_unit_8ornot.disjuncts[0],
            ],
        ]
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.WARNING):
            results = SolverFactory('gdpopt.loa').solve(
                eight_process,
                init_algorithm='custom_disjuncts',
                custom_init_disjuncts=initialize,
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
            )

            self.assertIn(
                "The following disjuncts from the custom disjunct "
                "initialization set number 0 were unused: goofy",
                output.getvalue().strip(),
            )
        # and the solution is optimal to boot
        self.assertTrue(fabs(value(eight_process.profit.expr) - 68) <= 1e-2)

    def test_LOA_custom_disjuncts(self):
        """Test logic-based OA with custom disjuncts initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        initialize = [
            # Use units 1, 4, 7, 8
            [
                eight_process.use_unit_1or2.disjuncts[0],
                eight_process.use_unit_3ornot.disjuncts[1],
                eight_process.use_unit_4or5ornot.disjuncts[0],
                eight_process.use_unit_6or7ornot.disjuncts[1],
                eight_process.use_unit_8ornot.disjuncts[0],
            ],
            # Use units 2, 4, 6, 8
            [
                eight_process.use_unit_1or2.disjuncts[1],
                eight_process.use_unit_3ornot.disjuncts[1],
                eight_process.use_unit_4or5ornot.disjuncts[0],
                eight_process.use_unit_6or7ornot.disjuncts[0],
                eight_process.use_unit_8ornot.disjuncts[0],
            ],
        ]

        def assert_correct_disjuncts_active(
            solver, subprob_util_block, discrete_problem_util_block
        ):
            iteration = solver.initialization_iteration
            discrete_problem = discrete_problem_util_block.model()
            subprob = subprob_util_block.model()
            if iteration >= 2:
                return  # only checking initialization
            disjs_should_be_active = initialize[iteration]
            seen = set()
            for orig_disj in disjs_should_be_active:
                parent_nm = orig_disj.parent_component().name
                idx = orig_disj.index()
                # Find the corresponding components on the discrete problem and
                # subproblem
                discrete_problem_parent = discrete_problem.component(parent_nm)
                subprob_parent = subprob.component(parent_nm)
                self.assertIsInstance(discrete_problem_parent, Disjunct)
                self.assertIsInstance(subprob_parent, Block)
                discrete_problem_disj = discrete_problem_parent[idx]
                subprob_disj = subprob_parent[idx]
                self.assertTrue(value(discrete_problem_disj.indicator_var))
                self.assertTrue(subprob_disj.active)
                seen.add(subprob_disj)
            # check that all the others are inactive (False)
            for disj in subprob_util_block.disjunct_list:
                if disj not in seen:
                    self.assertFalse(disj.active)

        SolverFactory('gdpopt.loa').solve(
            eight_process,
            init_algorithm='custom_disjuncts',
            custom_init_disjuncts=initialize,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            subproblem_initialization_method=assert_correct_disjuncts_active,
        )

        self.assertTrue(fabs(value(eight_process.profit.expr) - 68) <= 1e-2)

    @unittest.skipUnless(
        SolverFactory('appsi_gurobi').available(exception_flag=False)
        and SolverFactory('appsi_gurobi').license_is_valid(),
        "Legacy APPSI Gurobi solver is not available",
    )
    def test_auto_persistent_solver(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        m = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.loa').solve(m, mip_solver='appsi_gurobi')

        self.assertTrue(fabs(value(m.profit.expr) - 68) <= 1e-2)
        ct.check_8PP_solution(self, m, results)


@unittest.skipIf(
    not LOA_solvers_available,
    "Required subsolvers %s are not available" % (LOA_solvers,),
)
class TestGDPoptRIC(unittest.TestCase):
    """Tests for the GDPopt solver plugin."""

    def test_infeasible_GDP(self):
        """Test for infeasible GDP."""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 2))
        m.d = Disjunction(expr=[[m.x**2 >= 3, m.x >= 3], [m.x**2 <= -1, m.x <= -1]])
        m.o = Objective(expr=m.x)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.WARNING):
            SolverFactory('gdpopt.ric').solve(
                m, mip_solver=mip_solver, nlp_solver=nlp_solver
            )
            self.assertIn(
                "Set covering problem is infeasible.", output.getvalue().strip()
            )

    def test_GDP_nonlinear_objective(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=m.x**2)
        SolverFactory('gdpopt.ric').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertAlmostEqual(value(m.o), 0)

        m = ConcreteModel()
        m.x = Var(bounds=(-1, 10))
        m.y = Var(bounds=(2, 3))
        m.d = Disjunction(expr=[[m.x + m.y >= 5], [m.x - m.y <= 3]])
        m.o = Objective(expr=-m.x**2, sense=maximize)
        SolverFactory('gdpopt.ric').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertAlmostEqual(value(m.o), 0)

    def test_logical_constraints_on_disjuncts(self):
        m = models.makeLogicalConstraintsOnDisjuncts()
        SolverFactory('gdpopt.ric').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertAlmostEqual(value(m.x), 8)

    def test_boolean_vars_on_disjuncts(self):
        m = models.makeBooleanVarsOnDisjuncts()
        SolverFactory('gdpopt.ric').solve(
            m, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertAlmostEqual(value(m.x), 8)

    def test_RIC_8PP_default_init(self):
        """Test logic-based outer approximation with 8PP."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.ric').solve(
            eight_process, mip_solver=mip_solver, nlp_solver=nlp_solver, tee=False
        )
        ct.check_8PP_solution(self, eight_process, results)

    @unittest.skipUnless(
        license_available, "No BARON license--8PP logical problem exceeds demo size"
    )
    def test_RIC_8PP_logical_default_init(self):
        """Test logic-based outer approximation with 8PP."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_logical.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.ric').solve(
            eight_process, mip_solver=mip_solver, nlp_solver=nlp_solver, tee=False
        )
        ct.check_8PP_logical_solution(self, eight_process, results)

    @unittest.skipUnless(
        SolverFactory('gams').available(exception_flag=False),
        'GAMS solver not available',
    )
    def test_RIC_8PP_gams_solver(self):
        # Make sure that the duals are still correct
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.ric').solve(
            eight_process,
            mip_solver=mip_solver,
            nlp_solver='gams',
            max_slack=0,
            tee=False,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_RIC_8PP_force_NLP(self):
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.ric').solve(
            eight_process,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            force_subproblem_nlp=True,
            tee=False,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_RIC_strip_pack_default_init(self):
        """Test logic-based outer approximation with strip packing."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.ric').solve(
            strip_pack, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    def test_RIC_strip_pack_default_init_logical_constraints(self):
        """Test logic-based outer approximation with strip packing with
        logical constraints."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        # add logical constraints
        strip_pack.Rec3AboveOrBelowRec1 = LogicalConstraint(
            expr=strip_pack.no_overlap[1, 3]
            .disjuncts[2]
            .indicator_var.lor(strip_pack.no_overlap[1, 3].disjuncts[3].indicator_var)
        )
        strip_pack.Rec3RightOrLeftOfRec2 = LogicalConstraint(
            expr=strip_pack.no_overlap[2, 3]
            .disjuncts[0]
            .indicator_var.lor(strip_pack.no_overlap[2, 3].disjuncts[1].indicator_var)
        )
        SolverFactory('gdpopt.ric').solve(
            strip_pack, mip_solver=mip_solver, nlp_solver=nlp_solver
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 13) <= 1e-2)

    @unittest.pytest.mark.expensive
    def test_RIC_constrained_layout_default_init(self):
        """Test RIC with constrained layout."""
        exfile = import_file(join(exdir, 'constrained_layout', 'cons_layout_model.py'))
        cons_layout = exfile.build_constrained_layout_model()
        SolverFactory('gdpopt.ric').solve(
            cons_layout,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            iterlim=120,
            max_slack=5,  # problem is convex, so can decrease slack
        )
        objective_value = value(cons_layout.min_dist_cost.expr)
        self.assertTrue(
            fabs(objective_value - 41573) <= 200,
            "Objective value of %s instead of 41573" % objective_value,
        )

    def test_RIC_8PP_maxBinary(self):
        """Test logic-based OA with max_binary initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.ric').solve(
            eight_process,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_RIC_strip_pack_maxBinary(self):
        """Test RIC with strip packing using max_binary initialization."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.ric').solve(
            strip_pack,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    def test_RIC_strip_pack_maxBinary_logical_constraints(self):
        """Test RIC with strip packing using max_binary initialization and
        including logical constraints."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        # add logical constraints
        strip_pack.Rec3AboveOrBelowRec1 = LogicalConstraint(
            expr=strip_pack.no_overlap[1, 3]
            .disjuncts[2]
            .indicator_var.lor(strip_pack.no_overlap[1, 3].disjuncts[3].indicator_var)
        )
        strip_pack.Rec3RightOrLeftOfRec2 = LogicalConstraint(
            expr=strip_pack.no_overlap[2, 3]
            .disjuncts[0]
            .indicator_var.lor(strip_pack.no_overlap[2, 3].disjuncts[1].indicator_var)
        )
        SolverFactory('gdpopt.ric').solve(
            strip_pack,
            init_algorithm='max_binary',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 13) <= 1e-2)

    def test_RIC_8PP_fixed_disjuncts(self):
        """Test RIC with 8PP using fixed disjuncts initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        initialize = [
            # Use units 1, 4, 7, 8
            eight_process.use_unit_1or2.disjuncts[0],
            eight_process.use_unit_3ornot.disjuncts[1],
            eight_process.use_unit_4or5ornot.disjuncts[0],
            eight_process.use_unit_6or7ornot.disjuncts[1],
            eight_process.use_unit_8ornot.disjuncts[0],
        ]
        for disj in eight_process.component_data_objects(Disjunct):
            if disj in initialize:
                disj.binary_indicator_var.set_value(1)
            else:
                disj.binary_indicator_var.set_value(0)
        results = SolverFactory('gdpopt.ric').solve(
            eight_process,
            init_algorithm='fix_disjuncts',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
        )
        ct.check_8PP_solution(self, eight_process, results)

    def test_RIC_custom_disjuncts(self):
        """Test logic-based OA with custom disjuncts initialization."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        initialize = [
            # Use units 1, 4, 7, 8
            [
                eight_process.use_unit_1or2.disjuncts[0],
                eight_process.use_unit_3ornot.disjuncts[1],
                eight_process.use_unit_4or5ornot.disjuncts[0],
                eight_process.use_unit_6or7ornot.disjuncts[1],
                eight_process.use_unit_8ornot.disjuncts[0],
            ],
            # Use units 2, 4, 6, 8
            [
                eight_process.use_unit_1or2.disjuncts[1],
                eight_process.use_unit_3ornot.disjuncts[1],
                eight_process.use_unit_4or5ornot.disjuncts[0],
                eight_process.use_unit_6or7ornot.disjuncts[0],
                eight_process.use_unit_8ornot.disjuncts[0],
            ],
        ]

        def assert_correct_disjuncts_active(
            solver, subprob_util_block, discrete_problem_util_block
        ):
            # I can get the iteration based on the number of no-good
            # cuts in this case...
            iteration = solver.initialization_iteration
            discrete_problem = discrete_problem_util_block.model()
            subprob = subprob_util_block.model()
            if iteration >= 2:
                return  # only checking initialization
            disjs_should_be_active = initialize[iteration]
            seen = set()
            for orig_disj in disjs_should_be_active:
                parent_nm = orig_disj.parent_component().name
                idx = orig_disj.index()
                # Find the corresponding components on the discrete problem and
                # subproblem
                discrete_problem_parent = discrete_problem.component(parent_nm)
                subprob_parent = subprob.component(parent_nm)
                self.assertIsInstance(discrete_problem_parent, Disjunct)
                self.assertIsInstance(subprob_parent, Block)
                discrete_problem_disj = discrete_problem_parent[idx]
                subprob_disj = subprob_parent[idx]
                self.assertTrue(value(discrete_problem_disj.indicator_var))
                self.assertTrue(subprob_disj.active)
                seen.add(subprob_disj)
            # check that all the others are inactive (False)
            for disj in subprob_util_block.disjunct_list:
                if disj not in seen:
                    self.assertFalse(disj.active)

        SolverFactory('gdpopt.ric').solve(
            eight_process,
            init_algorithm='custom_disjuncts',
            custom_init_disjuncts=initialize,
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            subproblem_initialization_method=assert_correct_disjuncts_active,
        )

        self.assertTrue(fabs(value(eight_process.profit.expr) - 68) <= 1e-2)

    def test_force_nlp_subproblem_with_general_integer_variables(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers, bounds=(0, 10))
        m.y = Var(bounds=(0, 10))
        m.disjunction = Disjunction(
            expr=[
                [m.x**2 <= 4, m.y**2 <= 1],
                [(m.x - 1) ** 2 + (m.y - 1) ** 2 <= 4, m.y <= 4],
            ]
        )
        m.obj = Objective(expr=-m.y - m.x)
        results = SolverFactory('gdpopt.ric').solve(
            m,
            init_algorithm='no_init',
            mip_solver=mip_solver,
            nlp_solver=nlp_solver,
            force_subproblem_nlp=True,
        )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(value(m.x), 2)
        self.assertAlmostEqual(value(m.y), 1 + sqrt(3))

    # Put unbounded vars in the global constraints and make sure this gets
    # handled correctly by GDPopt (The GDP transformations will throw a fit if
    # we put unbounded variables on the Disjuncts, but you can put them
    # globally... Seems like a terrible idea, but as long as the behavior is
    # correct, I guess people can learn the hard way)
    def test_force_nlp_subproblem_with_unbounded_integer_variables(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers, bounds=(0, 10))
        m.y = Var(bounds=(0, 10))
        m.w = Var(domain=Integers)
        m.disjunction = Disjunction(
            expr=[
                [m.x**2 <= 4, m.y**2 <= 1],
                [(m.x - 1) ** 2 + (m.y - 1) ** 2 <= 4, m.y <= 4],
            ]
        )
        m.c = Constraint(expr=m.x + m.y == m.w)
        m.obj = Objective(expr=-m.w)
        # We don't find a feasible solution this way, but we want to make sure
        # we do a few iterations, and then handle the termination state
        # correctly.
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            results = SolverFactory('gdpopt.ric').solve(
                m,
                init_algorithm='no_init',
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
                force_subproblem_nlp=True,
                iterlim=5,
            )
        self.assertIn("No feasible solutions found.", output.getvalue().strip())
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxIterations
        )
        # There's no solution in the model
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.y.value)
        self.assertIsNone(m.w.value)
        self.assertIsNone(m.disjunction.disjuncts[0].indicator_var.value)
        self.assertIsNone(m.disjunction.disjuncts[1].indicator_var.value)
        self.assertIsNotNone(results.solver.user_time)


@unittest.skipIf(
    not GLOA_solvers_available,
    "Required subsolvers %s are not available" % (GLOA_solvers,),
)
@unittest.skipIf(not mcpp_available(), "MC++ is not available")
class TestGLOA(unittest.TestCase):
    """Tests for global logic-based outer approximation."""

    def test_GDP_integer_vars(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 10))
        m.y = Var(domain=Integers, bounds=(0, 5))
        m.d = Disjunction(expr=[[m.x >= m.y, m.y >= 3.5], [m.x >= m.y, m.y >= 2.5]])
        m.o = Objective(expr=m.x)
        SolverFactory('gdpopt').solve(
            m,
            algorithm='GLOA',
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            minlp_solver=minlp_solver,
        )
        self.assertAlmostEqual(value(m.o.expr), 3)

    def make_nonlinear_gdp_with_int_vars(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 10))
        m.y = Var(domain=Integers, bounds=(0, 5))
        m.d = Disjunction(
            expr=[[m.x**2 >= m.y, m.y >= 3.5], [m.x**2 >= m.y, m.y >= 2.5]]
        )
        m.o = Objective(expr=m.x)
        return m

    def test_nonlinear_GDP_integer_vars(self):
        m = self.make_nonlinear_gdp_with_int_vars()
        SolverFactory('gdpopt.gloa').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            minlp_solver=minlp_solver,
        )
        self.assertAlmostEqual(value(m.o.expr), sqrt(3))
        self.assertAlmostEqual(value(m.y), 3)

    def test_nonlinear_GDP_integer_vars_force_nlp_subproblem(self):
        m = self.make_nonlinear_gdp_with_int_vars()
        SolverFactory('gdpopt.gloa').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            minlp_solver=minlp_solver,
            force_subproblem_nlp=True,
        )
        self.assertAlmostEqual(value(m.o.expr), sqrt(3))
        self.assertAlmostEqual(value(m.y), 3)

    def test_GDP_integer_vars_infeasible(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.y = Var(domain=Integers, bounds=(0, 5))
        m.d = Disjunction(expr=[[m.x >= m.y, m.y >= 3.5], [m.x >= m.y, m.y >= 2.5]])
        m.o = Objective(expr=m.x)
        res = SolverFactory('gdpopt.gloa').solve(
            m,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            minlp_solver=minlp_solver,
        )
        self.assertEqual(
            res.solver.termination_condition, TerminationCondition.infeasible
        )

    @unittest.skipUnless(license_available, "Global NLP solver license not available")
    def test_logical_constraints_on_disjuncts(self):
        m = models.makeLogicalConstraintsOnDisjuncts()
        SolverFactory('gdpopt.gloa').solve(
            m, mip_solver=mip_solver, nlp_solver=global_nlp_solver
        )
        self.assertAlmostEqual(value(m.x), 8)

    @unittest.skipUnless(license_available, "Global NLP solver license not available")
    def test_boolean_vars_on_disjuncts(self):
        m = models.makeBooleanVarsOnDisjuncts()
        SolverFactory('gdpopt.gloa').solve(
            m, mip_solver=mip_solver, nlp_solver=global_nlp_solver
        )
        self.assertAlmostEqual(value(m.x), 8)

    @unittest.skipUnless(license_available, "Global NLP solver license not available.")
    def test_GLOA_8PP(self):
        """Test the global logic-based outer approximation algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.gloa').solve(
            eight_process,
            tee=False,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
        )
        ct.check_8PP_solution(self, eight_process, results)

    @unittest.skipUnless(license_available, "Global NLP solver license not available")
    def test_GLOA_8PP_logical(self):
        """Test the global logic-based outer approximation algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_logical.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.gloa').solve(
            eight_process,
            tee=False,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
        )
        ct.check_8PP_logical_solution(self, eight_process, results)

    @unittest.skipUnless(license_available, "Global NLP solver license not available.")
    def test_GLOA_8PP_force_NLP(self):
        """Test the global logic-based outer approximation algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.gloa').solve(
            eight_process,
            tee=False,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
            force_subproblem_nlp=True,
        )
        ct.check_8PP_solution(self, eight_process, results)

    @unittest.skipUnless(license_available, "Global NLP solver license not available.")
    def test_GLOA_strip_pack_default_init(self):
        """Test logic-based outer approximation with strip packing."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.gloa').solve(
            strip_pack,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    @unittest.skipUnless(license_available, "Global NLP solver license not available")
    def test_GLOA_strip_pack_default_init_logical_constraints(self):
        """Test logic-based outer approximation with strip packing."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        # add logical constraints
        strip_pack.Rec3AboveOrBelowRec1 = LogicalConstraint(
            expr=strip_pack.no_overlap[1, 3]
            .disjuncts[2]
            .indicator_var.lor(strip_pack.no_overlap[1, 3].disjuncts[3].indicator_var)
        )
        strip_pack.Rec3RightOrLeftOfRec2 = LogicalConstraint(
            expr=strip_pack.no_overlap[2, 3]
            .disjuncts[0]
            .indicator_var.lor(strip_pack.no_overlap[2, 3].disjuncts[1].indicator_var)
        )
        SolverFactory('gdpopt.gloa').solve(
            strip_pack,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 13) <= 1e-2)

    @unittest.skipUnless(license_available, "Global NLP solver license not available.")
    @unittest.pytest.mark.expensive
    def test_GLOA_constrained_layout_default_init(self):
        """Test LOA with constrained layout."""
        exfile = import_file(join(exdir, 'constrained_layout', 'cons_layout_model.py'))
        cons_layout = exfile.build_constrained_layout_model()
        results = SolverFactory('gdpopt.gloa').solve(
            cons_layout,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
            tee=False,
        )
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        objective_value = value(cons_layout.min_dist_cost.expr)
        self.assertTrue(
            fabs(objective_value - 41573) <= 200,
            "Objective value of %s instead of 41573" % objective_value,
        )

    def test_GLOA_ex_633_trespalacios(self):
        """Test LOA with Francisco thesis example."""
        exfile = import_file(join(exdir, 'small_lit', 'ex_633_trespalacios.py'))
        model = exfile.build_simple_nonconvex_gdp()
        SolverFactory('gdpopt.gloa').solve(
            model,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
            tee=False,
        )
        objective_value = value(model.obj.expr)
        self.assertAlmostEqual(objective_value, 4.46, 2)

    @unittest.skipUnless(license_available, "Global NLP solver license not available.")
    def test_GLOA_nonconvex_HENS(self):
        exfile = import_file(join(exdir, 'small_lit', 'nonconvex_HEN.py'))
        model = exfile.build_gdp_model()
        SolverFactory('gdpopt.gloa').solve(
            model,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
            tee=False,
        )
        objective_value = value(model.objective.expr)
        self.assertAlmostEqual(objective_value * 1e-5, 1.14385, 2)

    @unittest.skipUnless(license_available, "Global NLP solver license not available.")
    def test_GLOA_disjunctive_bounds(self):
        exfile = import_file(join(exdir, 'small_lit', 'nonconvex_HEN.py'))
        model = exfile.build_gdp_model()
        SolverFactory('gdpopt.gloa').solve(
            model,
            mip_solver=mip_solver,
            nlp_solver=global_nlp_solver,
            nlp_solver_args=global_nlp_solver_args,
            calc_disjunctive_bounds=True,
            tee=False,
        )
        objective_value = value(model.objective.expr)
        self.assertAlmostEqual(objective_value * 1e-5, 1.14385, 2)


@unittest.skipIf(
    not GLOA_solvers_available,
    "Required subsolvers %s are not available" % (GLOA_solvers,),
)
@unittest.skipIf(
    not LOA_solvers_available,
    "Required subsolvers %s are not available" % (LOA_solvers,),
)
class TestConfigOptions(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-10, 10))
        m.y = Var(bounds=(-10, 10))

        m.disjunction = Disjunction(
            expr=[[m.y >= m.x**2 + 1], [m.y >= m.x**2 - 3 * m.x + 2]]
        )

        m.obj = Objective(expr=m.y)

        return m

    @unittest.skipIf(not mcpp_available(), "MC++ is not available")
    @unittest.skipUnless(gurobi_available, 'Gurobi not available')
    def test_set_options_on_config_block(self):
        m = self.make_model()

        opt = SolverFactory('gdpopt.loa')
        with self.assertRaisesRegex(
            ValueError,
            r"Changing the algorithm in the solve method "
            r"is not supported for algorithm-specific "
            r"GDPopt solvers. Either use "
            r"SolverFactory[(]'gdpopt'[)] or instantiate a "
            r"solver with the algorithm you want to use.",
        ):
            opt.solve(m, algorithm='RIC')

        opt.CONFIG.mip_solver = mip_solver
        opt.CONFIG.nlp_solver = nlp_solver
        opt.CONFIG.init_algorithm = 'no_init'

        buf = StringIO()
        with redirect_stdout(buf):
            opt.solve(m, tee=True)
        self.assertIn('mip_solver: gurobi', buf.getvalue())
        self.assertAlmostEqual(value(m.obj), -0.25)

        opt = SolverFactory('gdpopt.loa')
        opt.config.mip_solver = 'glpk'
        # Make sure this didn't change the other options
        self.assertEqual(opt.CONFIG.mip_solver, mip_solver)
        self.assertEqual(opt.CONFIG.nlp_solver, nlp_solver)
        self.assertEqual(opt.CONFIG.init_algorithm, 'no_init')

        buf = StringIO()
        with redirect_stdout(buf):
            opt.solve(m, tee=True)
        self.assertIn('mip_solver: glpk', buf.getvalue())
        self.assertAlmostEqual(value(m.obj), -0.25)

    @unittest.skipIf(not mcpp_available(), "MC++ is not available")
    @unittest.skipUnless(gurobi_available, 'Gurobi not available')
    def test_set_options_in_init(self):
        m = self.make_model()

        opt = SolverFactory(
            'gdpopt.loa',
            mip_solver='gurobi',
            nlp_solver=nlp_solver,
            init_algorithm='no_init',
        )

        buf = StringIO()
        with redirect_stdout(buf):
            opt.solve(m, tee=True)
        self.assertIn('mip_solver: gurobi', buf.getvalue())
        self.assertAlmostEqual(value(m.obj), -0.25)

        buf = StringIO()
        with redirect_stdout(buf):
            opt.solve(m, tee=True, mip_solver='glpk')
        self.assertIn('mip_solver: glpk', buf.getvalue())
        self.assertAlmostEqual(value(m.obj), -0.25)
        self.assertEqual(opt.config.mip_solver, 'gurobi')

    @unittest.skipUnless(gurobi_available, 'Gurobi not available')
    def test_no_default_algorithm(self):
        m = self.make_model()

        opt = SolverFactory('gdpopt')

        buf = StringIO()
        with redirect_stdout(buf):
            opt.solve(
                m,
                algorithm='RIC',
                tee=True,
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
            )
        self.assertIn('using RIC algorithm', buf.getvalue())
        self.assertAlmostEqual(value(m.obj), -0.25)

        buf = StringIO()
        with redirect_stdout(buf):
            opt.solve(
                m,
                algorithm='LBB',
                tee=True,
                mip_solver=mip_solver,
                nlp_solver=nlp_solver,
                minlp_solver='gurobi',
            )
        self.assertIn('using LBB algorithm', buf.getvalue())
        self.assertAlmostEqual(value(m.obj), -0.25)


if __name__ == '__main__':
    unittest.main()
